@@ -13,38 +13,52 @@
 #include <unistd.h>
 #include <assert.h>
 
-static int sysinfo_updated;
 static struct sysinfo s_info;
 
 void prepare_update() {
-  sysinfo_updated = 0;
 }
 
 static void update_sysinfo() {
-  if (sysinfo_updated) return;
-  sysinfo_updated = 1;
-
   sysinfo(&s_info);
+
+  info.uptime = (double) s_info.uptime;
+
+  /* there was some problem with these */
+#if 0
+  info.loadavg[0] = s_info.loads[0] / 100000.0f;
+  info.loadavg[1] = s_info.loads[1] / 100000.0f;
+  info.loadavg[2] = s_info.loads[2] / 100000.0f;
+  info.mask |= 1 << INFO_LOADAVG;
+#endif
+
+  info.procs = s_info.procs;
+
+  /* these aren't nice, no cache and should check kernel version for mem_unit */
+#if 0
+  info.memmax = s_info.totalram;
+  info.mem = s_info.totalram - s_info.freeram;
+  info.swapmax = s_info.totalswap;
+  info.swap = s_info.totalswap - s_info.swap;
+  info.mask |= 1 << INFO_MEM;
+#endif
+
+  info.mask |= (1 << INFO_UPTIME) | (1 << INFO_PROCS);
 }
 
-double get_uptime() {
-  double t = 0;
-
+void update_uptime() {
   /* prefers sysinfo() for uptime, I don't really know which one is better
    * (=faster?) */
-
 #ifdef USE_PROC_UPTIME
   static int rep;
   FILE *fp = open_file("/proc/uptime", &rep);
   if(!fp) return 0;
-  fscanf(fp, "%lf", &t);
+  fscanf(fp, "%lf", &info.uptime);
   fclose(fp);
+
+  info.mask |= (1 << INFO_UPTIME);
 #else
   update_sysinfo();
-  t = (double) s_info.uptime;
 #endif
-
-  return t;
 }
 
 /* these things are also in sysinfo except Buffers:, that's why I'm reading
@@ -57,7 +71,8 @@ void update_meminfo() {
 /*  unsigned int a; */
   char buf[256];
 
-  mem = memmax = swap = swapmax = bufmem = buffers = cached = 0;
+  info.mem = info.memmax = info.swap = info.swapmax = info.bufmem =
+    info.buffers = info.cached = 0;
 
   if (meminfo_fp == NULL)
     meminfo_fp = open_file("/proc/meminfo", &rep);
@@ -69,33 +84,31 @@ void update_meminfo() {
     if (fgets(buf, 255, meminfo_fp) == NULL) break;
 
     if (strncmp(buf, "MemTotal:", 9) == 0) {
-      sscanf(buf, "%*s %u", &memmax);
+      sscanf(buf, "%*s %u", &info.memmax);
     }
     else if (strncmp(buf, "MemFree:", 8) == 0) {
-      sscanf(buf, "%*s %u", &mem);
+      sscanf(buf, "%*s %u", &info.mem);
     }
     else if (strncmp(buf, "SwapTotal:", 10) == 0) {
-      sscanf(buf, "%*s %u", &swapmax);
+      sscanf(buf, "%*s %u", &info.swapmax);
     }
     else if (strncmp(buf, "SwapFree:", 9) == 0) {
-      sscanf(buf, "%*s %u", &swap);
+      sscanf(buf, "%*s %u", &info.swap);
     }
     else if (strncmp(buf, "Buffers:", 8) == 0) {
-      sscanf(buf, "%*s %u", &buffers);
-      bufmem += buffers;
+      sscanf(buf, "%*s %u", &info.buffers);
     }
     else if (strncmp(buf, "Cached:", 7) == 0) {
-      sscanf(buf, "%*s %u", &cached);
-      bufmem += cached;
+      sscanf(buf, "%*s %u", &info.cached);
     }
   }
 
-  /* fclose(fp); meminfo_fp = NULL; */
+  info.mem = info.memmax - info.mem;
+  info.swap = info.swapmax - info.swap;
 
-  mem = memmax - mem;
-  swap = swapmax - swap;
+  info.bufmem = info.cached + info.buffers;
 
-  if (no_buffers) mem -= bufmem;
+  info.mask |= (1 << INFO_MEM) | (1 << INFO_BUFFERS);
 }
 
 static FILE *net_dev_fp;
@@ -138,6 +151,7 @@ void update_net_stats() {
     p++;
 
     ns = get_net_stat(s);
+    ns->up = 1;
     last_recv = ns->recv;
     last_trans = ns->trans;
 
@@ -168,13 +182,13 @@ void update_net_stats() {
   /* fclose(net_dev_fp); net_dev_fp = NULL; */
 }
 
-int get_total_processes() {
+void update_total_processes() {
   update_sysinfo();
-  return s_info.procs;
 }
 
-static int running_processes;
 static unsigned int cpu_user, cpu_system, cpu_nice;
+static double last_cpu_sum;
+static int clock_ticks;
 
 static FILE *stat_fp;
 
@@ -188,40 +202,71 @@ static void update_stat() {
     fseek(stat_fp, 0, SEEK_SET);
   if (stat_fp == NULL) return;
 
+  info.cpu_count = 0;
+
   while (!feof(stat_fp)) {
     if (fgets(buf, 255, stat_fp) == NULL)
       break;
 
     if (strncmp(buf, "procs_running ", 14) == 0) {
-      sscanf(buf, "%*s %d", &running_processes);
+      sscanf(buf, "%*s %d", &info.run_procs);
+      info.mask |= (1 << INFO_RUN_PROCS);
     }
     else if (strncmp(buf, "cpu ", 4) == 0) {
       sscanf(buf, "%*s %u %u %u", &cpu_user, &cpu_nice, &cpu_system);
+      info.mask |= (1 << INFO_CPU);
+    }
+    else if (strncmp(buf, "cpu", 3) == 0 && isdigit(buf[3])) {
+      info.cpu_count++;
     }
   }
 
-  /* fclose(fp); */
+  {
+    double delta;
+    delta = current_update_time - last_update_time;
+    if (delta <= 0.001) return;
+
+    if (clock_ticks == 0)
+      clock_ticks = sysconf(_SC_CLK_TCK);
+
+    info.cpu_usage = (cpu_user+cpu_nice+cpu_system - last_cpu_sum) / delta
+      / (double) clock_ticks / info.cpu_count;
+    last_cpu_sum = cpu_user+cpu_nice+cpu_system;
+  }
 }
 
-int get_running_processes() {
+void update_running_processes() {
   update_stat();
-  return running_processes;
 }
 
-static double last_cpu_sum;
-
 void update_cpu_usage() {
-  double delta;
-
-  delta = current_update_time - last_update_time;
-  if(delta <= 0.001) return;
-
   update_stat();
-
-  cpu_usage = (cpu_user+cpu_nice+cpu_system - last_cpu_sum) / delta / 100.0;
-  last_cpu_sum = cpu_user+cpu_nice+cpu_system;
 }
 
+void update_load_average() {
+#ifdef HAVE_GETLOADAVG
+  double v[3];
+  getloadavg(v, 3);
+  info.loadavg[0] = (float) v[0];
+  info.loadavg[1] = (float) v[1];
+  info.loadavg[2] = (float) v[2];
+#else
+  static int rep;
+  FILE *fp;
+
+  fp = open_file("/proc/loadavg", &rep);
+  if (!fp) {
+    v[0] = v[1] = v[2] = 0.0;
+    return;
+  }
+
+  fscanf(fp, "%f %f %f", &info.loadavg[0], &info.loadavg[1], &info.loadavg[2]);
+
+  fclose(fp);
+#endif
+}
+
+
 static int no_dots(const struct dirent *d) {
   if(d->d_name[0] == '.') return 0;
   return 1;
@@ -259,6 +304,7 @@ int open_i2c_sensor(const char *dev, const char *type, int n, int *div) {
   char path[256];
   char buf[64];
   int fd;
+  int divfd;
 
   /* if i2c device is NULL or *, get first */
   if (dev == NULL || strcmp(dev, "*") == 0) {
@@ -283,6 +329,23 @@ int open_i2c_sensor(const char *dev, const char *type, int n, int *div) {
   else
     *div = 0;
 
+  /* test if *_div file exist, open it and use it as divisor */
+  snprintf(path, 255, I2C_DIR "%s/%s%d_div", dev, type, n);
+
+  divfd = open(path, O_RDONLY);
+  if (divfd > 0) {
+    /* read integer */
+    char divbuf[64];
+    unsigned int divn;
+    divn = read(divfd, divbuf, 63);
+    /* should read until n == 0 but I doubt that kernel will give these
+     * in multiple pieces. :) */
+    divbuf[divn] = '\0';
+    *div = atoi(divbuf);
+  }
+
+  close(divfd);
+
   return fd;
 }
 
@@ -305,37 +368,92 @@ double get_i2c_info(int fd, int div) {
   }
 
   /* divide voltage and temperature by 1000 */
-  if (div)
+  /* or if any other divisor is given, use that */ 
+  if (div > 1)
+    return val / div;
+  else if (div)
     return val / 1000.0;
   else
     return val;
 }
 
-void get_load_average(double v[3]) {
-#ifdef HAVE_GETLOADAVG
-  getloadavg(v, 3);
-#else
+#define ADT746X_FAN "/sys/devices/temperatures/cpu_fan_speed"
+
+static char *adt746x_fan_state;
+
+char* get_adt746x_fan() {
   static int rep;
   FILE *fp;
 
-  fp = open_file("/proc/loadavg", &rep);
-  if (!fp) {
-    v[0] = v[1] = v[2] = 0.0;
-    return;
+  if (adt746x_fan_state == NULL) {
+    adt746x_fan_state = (char*)malloc(100);
+    assert(adt746x_fan_state!=NULL);
   }
 
-  fscanf(fp, "%f %f %f", &v[0], &v[1], &v[2]);
-
+  fp = open_file(ADT746X_FAN, &rep);
+  if (!fp) {
+    strcpy(adt746x_fan_state,"No fan found! Hey, you don't have one?");
+    return adt746x_fan_state;
+  }
+  fscanf(fp, "%s", adt746x_fan_state);
   fclose(fp);
-#endif
+
+  return adt746x_fan_state;
 }
 
-#define ACPI_FAN "/proc/acpi/fan/FAN/state"
+#define ADT746X_CPU "/sys/devices/temperatures/cpu_temperature"
+
+static char *adt746x_cpu_state;
+
+char* get_adt746x_cpu() {
+  static int rep;
+  FILE *fp;
+
+  if (adt746x_cpu_state == NULL) {
+    adt746x_cpu_state = (char*)malloc(100);
+    assert(adt746x_cpu_state!=NULL);
+  }
+
+  fp = open_file(ADT746X_CPU, &rep);
+  fscanf(fp, "%2s", adt746x_cpu_state);
+  fclose(fp);
+
+  return adt746x_cpu_state;
+}
+
+static char *frequency;
+
+char* get_freq()
+{
+    FILE *f;
+    char s[1000];
+    if (frequency == NULL) {
+    	frequency = (char*)malloc(100);
+    	assert(frequency!=NULL);
+    }
+    //char frequency[10];
+    f=fopen("/proc/cpuinfo","r"); //open the CPU information file
+    //if (!f)
+    //    return;
+    while (fgets(s,1000,f)!=NULL)  //read the file
+        if ( strncmp(s, "cpu M", 5)==0 ) { //and search for the cpu mhz
+                //printf("%s", strchr(s, ':')+2);
+                strcpy(frequency,strchr(s, ':')+2); //copy just the number
+                frequency[strlen(frequency)-1] = '\0'; // strip \n
+                break;
+        }
+    fclose(f);
+    return frequency;
+}
+
+#define ACPI_FAN_DIR "/proc/acpi/fan/"
 
 static char *acpi_fan_state;
 
 char* get_acpi_fan() {
   static int rep;
+  char buf[256];
+  char buf2[256];
   FILE *fp;
 
   if (acpi_fan_state == NULL) {
@@ -343,9 +461,15 @@ char* get_acpi_fan() {
     assert(acpi_fan_state!=NULL);
   }
 
-   fp = open_file(ACPI_FAN, &rep);
+  /* yeah, slow... :/ */
+  if (!get_first_file_in_a_directory(ACPI_FAN_DIR, buf, &rep))
+    return "no fans?";
+
+  snprintf(buf2, 256, "%s%s/state", ACPI_FAN_DIR, buf);
+
+  fp = open_file(buf2, &rep);
   if (!fp) {
-    strcpy(acpi_fan_state,"No fan found! Hey, you don't have one?");
+    strcpy(acpi_fan_state, "can't open fan's state file");
     return acpi_fan_state;
   }
   fscanf(fp, "%*s %99s", acpi_fan_state);
@@ -353,12 +477,14 @@ char* get_acpi_fan() {
   return acpi_fan_state;
 }
 
-#define ACPI_AC_ADAPTER "/proc/acpi/ac_adapter/ACAD/state"
+#define ACPI_AC_ADAPTER_DIR "/proc/acpi/ac_adapter/"
 
 static char *acpi_ac_adapter_state;
 
 char* get_acpi_ac_adapter() {
   static int rep;
+  char buf[256];
+  char buf2[256];
   FILE *fp;
 
   if (acpi_ac_adapter_state == NULL) {
@@ -366,7 +492,13 @@ char* get_acpi_ac_adapter() {
     assert(acpi_ac_adapter_state!=NULL);
   }
 
-  fp = open_file(ACPI_AC_ADAPTER, &rep);
+  /* yeah, slow... :/ */
+  if (!get_first_file_in_a_directory(ACPI_AC_ADAPTER_DIR, buf, &rep))
+    return "no ac_adapters?";
+
+  snprintf(buf2, 256, "%s%s/state", ACPI_AC_ADAPTER_DIR, buf);
+
+  fp = open_file(buf2, &rep);
   if (!fp) {
     strcpy(acpi_ac_adapter_state,"No ac adapter found.... where is it?");
     return acpi_ac_adapter_state;
@@ -414,13 +546,22 @@ int open_acpi_temperature(const char *name) {
   return fd;
 }
 
-double get_acpi_temperature(int fd) {
-  double temp;
+static double last_acpi_temp;
+static double last_acpi_temp_time;
 
+double get_acpi_temperature(int fd) {
   if (fd <= 0) return 0;
 
+  /* don't update acpi temperature too often */
+  if (current_update_time - last_acpi_temp_time < 11.32) {
+    return last_acpi_temp;
+  }
+  last_acpi_temp_time = current_update_time;
+
+  /* seek to beginning */
   lseek(fd, 0, SEEK_SET);
 
+  /* read */
   {
     char buf[256];
     int n;
@@ -428,11 +569,11 @@ double get_acpi_temperature(int fd) {
     if (n < 0) ERR("can't read fd %d: %s", fd, strerror(errno));
     else {
       buf[n] = '\0';
-      sscanf(buf, "temperature: %lf", &temp);
+      sscanf(buf, "temperature: %lf", &last_acpi_temp);
     }
   }
 
-  return temp;
+  return last_acpi_temp;
 }
 
 /*
@@ -488,10 +629,10 @@ static char last_battery_str[64];
 
 static double last_battery_time;
 
-void get_battery_stuff(char *buf, unsigned int n, int b) {
+void get_battery_stuff(char *buf, unsigned int n, const char *bat) {
   static int rep, rep2;
   char acpi_path[128];
-  snprintf(acpi_path, 127, ACPI_BATTERY_BASE_PATH "/BAT%d/state", b);
+  snprintf(acpi_path, 127, ACPI_BATTERY_BASE_PATH "/%s/state", bat);
 
   /* don't update battery too often */
   if (current_update_time - last_battery_time < 29.5) {
@@ -513,7 +654,10 @@ void get_battery_stuff(char *buf, unsigned int n, int b) {
     /* read last full capacity if it's zero */
     if (acpi_last_full == 0) {
       static int rep;
-      FILE *fp = open_file(acpi_path, &rep);
+      char path[128];
+      FILE *fp;
+      snprintf(path, 127, ACPI_BATTERY_BASE_PATH "/%s/info", bat);
+      fp = open_file(path, &rep);
       if (fp != NULL) {
         while (!feof(fp)) {
           char b[256];
@@ -536,28 +680,45 @@ void get_battery_stuff(char *buf, unsigned int n, int b) {
       if (fgets(buf, 256, acpi_bat_fp) == NULL) break;
 
       /* let's just hope units are ok */
-      sscanf(buf, "charging state: %63s", charging_state);
-      sscanf(buf, "present rate: %d", &present_rate);
-      sscanf(buf, "remaining capacity: %d", &remaining_capacity);
+      if(buf[0] == 'c')
+        sscanf(buf, "charging state: %63s", charging_state);
+      else if(buf[0] == 'p')
+        sscanf(buf, "present rate: %d", &present_rate);
+      else if(buf[0] == 'r')
+        sscanf(buf, "remaining capacity: %d", &remaining_capacity);
     }
 
-    if (present_rate <= 0 || remaining_capacity <= 0) {
-      strcpy(last_battery_str, "AC");
-    }
-    else {
-      if (strcmp(charging_state, "charging") == 0) {
-        if (acpi_last_full != 0) {
-          strcpy(last_battery_str, "charging ");
-          format_seconds(last_battery_str+9, 63-9,
-              (acpi_last_full - remaining_capacity) * 60 * 60 / present_rate);
-        }
-        else {
-          strcpy(last_battery_str, "charging");
-        }
+    /* charging */
+    if (strcmp(charging_state, "charging") == 0) {
+      if (acpi_last_full != 0 && present_rate > 0) {
+        strcpy(last_battery_str, "charging ");
+        format_seconds(last_battery_str+9, 63-9,
+            (acpi_last_full - remaining_capacity) * 60 * 60 / present_rate);
       }
-      else
+      else if (acpi_last_full != 0 && present_rate <= 0) {
+        sprintf(last_battery_str, "charging %d%%",
+            remaining_capacity * 100 / acpi_last_full);
+      }
+      else {
+        strcpy(last_battery_str, "charging");
+      }
+    }
+    /* discharging */
+    else if (strcmp(charging_state, "discharging") == 0) {
+      if (present_rate > 0)
         format_seconds(last_battery_str, 63,
             (remaining_capacity * 60 * 60) / present_rate);
+      else
+        sprintf(last_battery_str, "discharging %d%%",
+            remaining_capacity * 100 / acpi_last_full);
+    }
+    /* unknown, probably full / AC */
+    else {
+      if (acpi_last_full != 0 && remaining_capacity != acpi_last_full)
+        sprintf(last_battery_str, "unknown %d%%",
+            remaining_capacity * 100 / acpi_last_full);
+      else
+        strcpy(last_battery_str, "AC");
     }
   }
   else {
